/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Threading;

namespace VDS.Web.Logging
{
    /// <summary>
    /// An Apache style Logger which logs to files
    /// </summary>
    public class FileLogger
        : StreamLogger
    {
        private readonly String _logFile;

        /// <summary>
        /// Creates a new Logger
        /// </summary>
        /// <param name="logFile">Log File</param>
        public FileLogger(String logFile)
            : this(logFile, LogCommon) { }

        /// <summary>
        /// Creates a new Logger
        /// </summary>
        /// <param name="logFile">Log File</param>
        /// <param name="logFormatString">Log Format</param>
        public FileLogger(String logFile, String logFormatString)
            : base(FileLogger.OpenFileStream(logFile), logFormatString)
        {
            this._logFile = logFile;
        }

        /// <summary>
        /// Indicates that the logger should close the stream when the server transitions to a stopped/disposed state
        /// </summary>
        protected override bool ShouldClose
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Reopens the log file
        /// </summary>
        /// <returns>Log File stream</returns>
        protected override StreamWriter RestartStream()
        {
            return new StreamWriter(FileLogger.OpenFileStream(this._logFile));
        }

        /// <summary>
        /// Helper method for setting up the file stream
        /// </summary>
        /// <param name="logFile">Log File</param>
        /// <returns>Log File stream</returns>
        private static Stream OpenFileStream(String logFile)
        {
            if (logFile == null) throw new ArgumentNullException("logFile", "Log File cannot be null");
            try
            {
                if (!File.Exists(logFile))
                {
                    return File.Create(logFile);
                }
                else
                {
                    return File.Open(logFile, FileMode.Append, FileAccess.Write);
                }
            }
            catch (Exception ex)
            {
                throw new ArgumentException("Unable to create/access the Log File '" + logFile + "'", "logFile", ex);
            }
        }
    }
}
