/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Threading;

namespace VDS.Web.Logging
{
    /// <summary>
    /// An Apache style Logger which logs to a stream
    /// </summary>
    public class StreamLogger
        : ApacheStyleLogger
    {
        private StreamWriter _writer;
        private ReaderWriterLockSlim _lock = new ReaderWriterLockSlim();

        /// <summary>
        /// Creates a new logger
        /// </summary>
        /// <param name="writer">Stream Writer</param>
        /// <param name="logFormatString">Log Format</param>
        public StreamLogger(StreamWriter writer, String logFormatString)
            : base(logFormatString)
        {
            if (writer == null) throw new ArgumentNullException("Cannot create a Stream Logger that uses a null StreamWriter");
            this._writer = writer;
        }

        /// <summary>
        /// Creates a new logger
        /// </summary>
        /// <param name="writer">Stream Writer</param>
        public StreamLogger(StreamWriter writer)
            : this(writer, LogCommon) { }

        /// <summary>
        /// Creates a new logger
        /// </summary>
        /// <param name="stream">Stream</param>
        public StreamLogger(Stream stream)
            : this(new StreamWriter(stream)) { }

        /// <summary>
        /// Creates a new logger
        /// </summary>
        /// <param name="stream">Stream</param>
        /// <param name="logFormatString">Log Format</param>
        public StreamLogger(Stream stream, String logFormatString)
            : this(new StreamWriter(stream), logFormatString) { }

        /// <summary>
        /// Appends log lines to the stream
        /// </summary>
        /// <param name="line">Log line</param>
        protected override void AppendToLog(string line)
        {
            try
            {
                this._lock.EnterWriteLock();
                this._writer.WriteLine(line);
                this._writer.Flush();
            }
            finally
            {
                this._lock.ExitWriteLock();
            }
        }

        /// <summary>
        /// Gets whether the writer should be closed when the server transitions to Closed/Disposed state
        /// </summary>
        protected virtual bool ShouldClose
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Method which returns the stream writer to use
        /// </summary>
        /// <returns></returns>
        /// <remarks>
        /// Only invoked if the implementation indicates that it wants the writer to be closed when the server transitions to the Stopped/Disposed state
        /// </remarks>
        protected virtual StreamWriter RestartStream()
        {
            if (this._writer == null) throw new HttpServerException("StreamLogger does not support restarting over a server Stop state");
            return this._writer;
        }

        /// <summary>
        /// Handles the server post-state change notification
        /// </summary>
        /// <param name="server">HTTP Server</param>
        /// <remarks>
        /// If the <see cref="StreamLogger.ShouldClose"/> property has been overridden to return true and the server state is now Stopped/Disposed then the logger will close the writer
        /// </remarks>
        public sealed override void PostStateChange(HttpServer server)
        {
            base.PostStateChange(server);
            switch (server.State)
            {
                case ServerState.Stopped:
                case ServerState.Disposed:
                    if (this.ShouldClose)
                    {
                        try
                        {
                            this._lock.EnterWriteLock();
                            if (this._writer != null)
                            {
                                this._writer.Close();
                                this._writer = null;
                            }
                        }
                        finally
                        {
                            this._lock.ExitWriteLock();
                        }
                    }
                    break;
            }
            this.PostStateChangeInternal(server);
        }

        /// <summary>
        /// Method which derived classes may override to add their own post-state change logic
        /// </summary>
        /// <param name="server">HTTP Server</param>
        protected virtual void PostStateChangeInternal(HttpServer server)
        {
            //Do nothing
        }

        /// <summary>
        /// Disposes of the logger
        /// </summary>
        public override void Dispose()
        {
            if (this._writer != null)
            {
                this._writer.Close();
                this._writer = null;
            }
            base.Dispose();
        }
    }
}
