/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using VDS.Web.Handlers;

namespace VDS.Web
{
    [TestClass]
    public class ResponseTruncationTests
    {
        private readonly int TestPort = PortManager.GetNextPort();

        private HttpServer _server;
        private GenerateDataHandler _handler;

        [TestInitialize]
        public void Setup()
        {
            HttpListenerHandlerCollection handlers = new HttpListenerHandlerCollection();
            handlers.AddMapping(new HttpRequestMapping(HttpRequestMapping.AllVerbs, HttpRequestMapping.AnyPath, typeof(GenerateDataHandler)));
            this._server = new HttpServer(TestPort, handlers);
            this._server.Start();

            this._handler = this._server.Handlers.GetHandler(typeof(GenerateDataHandler)) as GenerateDataHandler;

            Assert.IsNotNull(this._handler);
        }

        [TestCleanup]
        public void Teardown()
        {
            this._server.Dispose();
        }

        private String ServerUrl
        {
            get
            {
                return "http://localhost:" + TestPort + "/test";
            }
        }

        private void RunTest(String data, int repeats)
        {
            String url = this.ServerUrl + "?data=" + Uri.EscapeDataString(data) + "&repeats=" + repeats;
            using (WebClient client = new WebClient())
            {
                String returnedData = client.DownloadString(url);
                Assert.AreEqual(data.Length * repeats, returnedData.Length);
            }
        }

        private void RunUnsafeTest(String data, int repeats)
        {
            String url = this.ServerUrl + "?data=" + Uri.EscapeDataString(data) + "&repeats=" + repeats + "&unsafe=true";
            using (WebClient client = new WebClient())
            {
                String returnedData = client.DownloadString(url);
                Assert.AreNotEqual(data.Length * repeats, returnedData.Length);
            }
        }

        [TestMethod]
        public void ResponseTruncation01()
        {
            this.RunTest("test", 1);
        }


        [TestMethod]
        public void ResponseTruncation02()
        {
            this.RunTest("test", 10);
        }

        [TestMethod]
        public void ResponseTruncation03()
        {
            this.RunTest("test", 100);
        }

        [TestMethod]
        public void ResponseTruncation04()
        {
            this.RunTest("test", 1000);
        }

        [TestMethod]
        public void ResponseTruncation05()
        {
            this.RunUnsafeTest("test", 1);
        }

        [TestMethod]
        public void ResponseTruncation06()
        {
            this.RunUnsafeTest("test", 10);
        }

        [TestMethod]
        public void ResponseTruncation07()
        {
            this.RunUnsafeTest("test", 100);
        }

        [TestMethod]
        public void ResponseTruncation08()
        {
            this.RunUnsafeTest("test", 1000);
        }
    }
}
